#include <Elasticity.h>
#include <Print.h>
#include <Behavior.h>
#include <Clock.h>
#include <Integration_runge.h>
#include <Integration_theta.h>
#include <Calcul_timer.h>
#include <Coefficient.h>
#include <Defines.h>

Z_START_NAMESPACE;                       //Debut du namespace Z_set

class NORTON_ISO_C : public BEHAVIOR, 
                      public THETA_INTEGRATOR,
                      public RUNGE_INTEGRATOR {    
 public : 
     AUTO_PTR<ELASTICITY>     elasticity;
     COEFF          K, n, R0, Q, b;
     TENSOR2_VINT         eel;
     SCALAR_VINT          evcum;
     TENSOR2_VAUX         evi;
     TENSOR2_FLUX         sig;
     TENSOR2_GRAD         eto;
     TENSOR2 deto;     
     AUTO_PTR<RUNGE>      runge_theta;
     AUTO_PTR<THETA>      theta_integ;
     MATRIX      m_tg_matrix;        // storage for the tangent matrix
  
     VECTOR      f_vec;              // residual 
     VECTOR      d_chi;              // increment of integrated vars
     VECTOR      f_0;                // what f_vec is compared to 
     SMATRIX     f_grad;             // the jacobian matrix
     VECTOR dgrad;  
  public :
    NORTON_ISO_C() {};
    virtual ~NORTON_ISO_C() {};

    virtual void initialize(ASCII_FILE&,int,LOCAL_INTEGRATION*);
    virtual void derivative(double,const  VECTOR&,VECTOR&);
    virtual void calc_grad_f(VECTOR&,SMATRIX&,const VECTOR&,const VECTOR&,double,double);
    INTEGRATION_RESULT* integrate(MAT_DATA&,const VECTOR& delta_grad,MATRIX*&,int);
};

Z_END_NAMESPACE;      //Fin du namespace Z_set
Z_USE_NAMESPACE;      //Utilisation du namespace Z_set

DECLARE_OBJECT(BEHAVIOR,NORTON_ISO_C,norton_iso_c)
 
void NORTON_ISO_C::initialize(ASCII_FILE& file, int dim, LOCAL_INTEGRATION* integ) 
{
  BEHAVIOR::initialize(file,dim,integ);  
  eel.initialize(this,"eel",tsz(),1); //tsz() :  problem (tensor) size sym 
  evcum.initialize(this,"evcum",1);
  evi.initialize(this,"evi",tsz(),1);
  sig.initialize(this,"sig",tsz(),1);
  eto.initialize(this,"eto",tsz(),1);
  
   for (;;) {
       STRING str=file.getSTRING();
       if (str.start_with("***") ) { file.back(); break; } 
       else if (str=="**model_coef") {
            str=file.getSTRING();
            while ((!str.start_with("*")) && file.ok) {
                if      (str=="K") K.read("K",file,this); 
                else if (str=="n") n.read("n",file,this); 
                else if (str=="R0") R0.read("R0",file,this); 
                else if (str=="Q") Q.read("Q",file,this); 
                else if (str=="b") b.read("b",file,this); 
                else {  INPUT_ERROR("Unknown coefficient: "+str); exit(1); }
                str=file.getSTRING();
            } file.back(); 
       }
       else if (str=="**elasticity" || str=="**ELASTICITY") {
            elasticity=ELASTICITY::read(file,this);
       } 
       else INPUT_ERROR("Unknown command: "+str); 
   }
   
    if (integ) {
      integration=integ->type();
      if (integration&LOCAL_INTEGRATION::RUNGE_ID) {
           runge_theta=(RUNGE*)integ;
           runge_theta->set_integrator((RUNGE_INTEGRATOR*)this);
      }
      else if (integration&LOCAL_INTEGRATION::THETA_ID) {
           theta_integ=(THETA*)integ;
           theta_integ->set_integrator((THETA_INTEGRATOR*)this);
      }
      else ERROR("Integration method is not allowed");
   }
   else {
      runge_theta=(RUNGE*)Create_object(LOCAL_INTEGRATION, "runge_kutta");
      runge_theta->initialize(1.e-3, 1.e-3);
      runge_theta->set_integrator((RUNGE_INTEGRATOR*)this);
      integration=runge_theta->type();
   }
  m_tg_matrix.resize(tsz()); m_tg_matrix=0.;
}

 
INTEGRATION_RESULT* NORTON_ISO_C::integrate(MAT_DATA& mdat,const VECTOR& delta_grad, MATRIX*& tg_matrix, int)
{ INTEGRATION_RESULT* ok =NULL;
  Timer_counter.start_time("Local Integration");
  tg_matrix = &m_tg_matrix;
  
  attach_all(mdat);
  calc_local_coefs();

  set_var_aux_to_var_aux_ini();
  set_var_int_to_var_int_ini();
  dgrad = delta_grad;
  deto.reassign(eto.size(),dgrad,0); 

   if (integration&LOCAL_INTEGRATION::THETA_ID) {
       VECTOR f_0=0.;
       calc_material(theta_integ->give_theta());
       SMATRIX f_grad(int_sz()); VECTOR d_chi(int_sz()), f_vec(int_sz());
       d_chi =0.;
       
       ok=theta_integ->theta_method(f_vec,f_grad,curr_mat_data->var_int(vint_index), d_chi,f_0,Time_ini,Dtime);
       if (ok) return ok;

       calc_material(1.0);
       VECTOR dvar = curr_mat_data->var_int(vint_index) - curr_mat_data->var_int_ini(vint_index); 
       attach(curr_mat_data->var_int(vint_index), dvar); 
       TENSOR4 tmp(tsz(),f_grad,0,0);
       if (Dtime>0.0) m_tg_matrix=*elasticity*tmp;
       else           m_tg_matrix=*elasticity;
  }
  else if (integration&LOCAL_INTEGRATION::RUNGE_ID) {
     if(Dtime!=0.) {
       dgrad/=Dtime;
       ok=runge_theta->runge_kutta(Time_ini,Dtime,curr_mat_data->var_int(vint_index));
     }
     if (CALC_TG_MATRIX) m_tg_matrix=*elasticity;
  }
  else  NOT_IMPLEMENTED_ERROR("Bad or unknown integration");

  evi=eto-eel; 
  sig = *elasticity*eel;

  Timer_counter.stop_time("Local Integration");
  Timer_counter.increment_count("in Local Int");
  return ok;
}


void NORTON_ISO_C::derivative(double tau, const VECTOR& chi, VECTOR& dchi)  
{
  calc_material((tau-Time_ini)/Dtime );
  attach((VECTOR&)chi);
  calc_local_coefs();
  dchi.resize(int_sz());
 
  sig = *elasticity*eel;

  TENSOR2 sprime = deviator(sig);
  TENSOR2 sigeff = sprime;
  double  J      = sqrt(1.5*(sigeff|sigeff));
  double  R      = R0 + Q*(1.-exp(-b*evcum));
  double  f      = J - R;

  TENSOR2 deto(tsz(),dgrad,0);
  TENSOR2 norm(tsz()); norm = 0.;
  TENSOR2  deel(tsz(),dchi,eel.start_pos);
  double& devcum = dchi[evcum.start_pos];

  if (f>0.0) {
     norm     = sigeff*(1.5/J);
     devcum   = pow(f/K,n);
  }  else {     devcum = 0.;     norm=0.; }

  deel = deto - devcum*norm;
}


void NORTON_ISO_C::calc_grad_f(VECTOR& f_vec, SMATRIX& f_grad,
                                const VECTOR& chi_vec,
                                const VECTOR& d_chi,
                                double theta, double dt)
{
   attach((VECTOR&)chi_vec, (VECTOR&)d_chi);
   calc_material(theta);


   TENSOR2  deel(tsz(),((VECTOR&)d_chi),eel.start_pos);
   double& devcum = ((VECTOR&)d_chi)[evcum.start_pos];
   
   VECTOR   f_vec_eel;
   f_vec_eel.reassign(eel.size(),f_vec,eel.start_pos);   
   double&  f_vec_evcum=f_vec[evcum.start_pos];
   
   SMATRIX  deel_deel;
   MATRIX   deel_devcum, devcum_deel;      
   deel_deel.reassign(eel.size(),f_grad,eel.start_pos,eel.start_pos);
   deel_devcum.reassign(eel.size(),1,f_grad,eel.start_pos,evcum.start_pos); 
   double&  devcum_devcum=f_grad(evcum.start_pos,evcum.start_pos);
   devcum_deel.reassign(1,eel.size(),f_grad,evcum.start_pos,eel.start_pos);    
     
    
    double tdt=theta*dt;
   f_grad=0.0; f_grad.add_to_diagonal(1.0);
   f_vec=d_chi;
   ELASTICITY& E=*elasticity;
   double tdv = theta*devcum;
   sig = E*eel;
   double       R   = R0+Q*(1-exp(-b*evcum));
   TENSOR2 sprime = deviator(sig);
   double  J      = sqrt(1.5*(sprime|sprime));
   double  f      = J - R;
   
   TENSOR4         unit, unit32;   // utility dev operator s' = unit*sigma 
   unit.resize(tsz());
   unit=0.0;
   double th=1.0/3.0;
   double th2=2.0/3.0;
   switch (tsz())  {
   case 6:
      unit(4,4)=unit(5,5)=1.0;
   case 4:
      unit(3,3)=1.0;
      unit(1,1)=unit(2,2)=th2;
      unit(1,0)=unit(0,1)=-th;
      unit(2,0)=unit(0,2)=-th;
      unit(2,1)=unit(1,2)=-th;
   case 1:
      unit(0,0)=th2;
   }
   unit32.resize(tsz()); unit32 = 1.5*unit; 
    
   f_vec_eel -= deto;
   if ( (f>0.0 && devcum>=0) || (devcum>0.0) ) {
       TENSOR2        norm            = sprime*(1.5/J);
       f_vec_eel      += norm*devcum;
       f_vec_evcum    -= dt*pow(f/K,n);
       SMATRIX dn_ds = unit32; dn_ds -= norm^norm; dn_ds *= tdv/J;
       SMATRIX dn_deel = dn_ds*E;
       double  dv_df   = tdt*n*pow(f/K,n-1)/K;
       TENSOR2 df_fs   = dv_df*norm;
       deel_deel         += dn_deel;
       deel_devcum        = norm;
       devcum_devcum     += dv_df*Q*b*exp(-b*evcum);
       devcum_deel       -= df_fs*E;
   }
} 
     
// ============================================================================ 
//   MATERIAL_PIECE        - calc_material just calls the sub_pieces...
//                         - coefficients in sub_pieces should only have 
//                           access to variables in its super MATERIAL_PIECE
//                         - put mamagement for calc_coef optimization into 
//                           this class?
// 
//  tsz             tensor size
//  calc_coef       calls calc_coef in sub_pieces by def. 
//  local_vars      a list automatically installed by 
//                  STORED_VARIABLE_SPEC.. 
// calc_material.   computes all coefficients interpolating external variables
//  attach_all      hooks on EVERYTHING
//  attach          links all local_vint to a temp vint vector.
//  init_name       these functions serve to line up the variables among
//                  a hierarchy of material bricks... start pos goes in, 
//                  end pos is returned
// ============================================================================ 
    
  
